/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * giounix.c: IO Channels using unix file descriptors
 * Copyright 1998 Owen Taylor
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GLib Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/. 
 */

/* 
 * MT safe
 */

#include "glib.h"
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#undef MIN
#undef MAX
#include <kernel/OS.h>
#include <stdlib.h>
/*
 * Beos IO Channels
 */

typedef struct _GIOBeosChannel GIOBeosChannel;
typedef struct _GIOBeosWatch GIOBeosWatch;

typedef enum {
  G_IO_WINDOWS_MESSAGES,	/* BMessages - not yet implemented!!!! */
  G_IO_FILE_DESC,		    /* Files/Pipes (as per unix)*/
  G_IO_SOCKET		        /* Sockets */
} GIOBeosChannelType;

struct _GIOBeosChannel {
  GIOChannel          channel;
  gint                fd;
  GIOBeosChannelType  type;
  char               *byte;
  int                 waiting;
  sem_id			  mutex;
};

/* The sock is used to provide an easy way of getting a unique number!
 * We create a sock and then add 1000 to it to give us a unique number
 * that can be keyed offf when a message arrives on our watch port.
 * This is only done for fd's and is totally bogus but was easy to
 * code and seems to work OK.  I didn't want to get involved in
 * maintaining a unique sequence and this does the job for us.
 */
struct _GIOBeosWatch {
  GPollFD       pollfd;
  GIOChannel   *channel;
  GIOCondition  condition;
  GIOFunc       callback;
  int			sock;
  thread_id     tid;
};

/* file/pipe definitions!!! */

static GIOError g_io_beos_read (GIOChannel *channel, 
		       gchar     *buf, 
		       guint      count,
		       guint     *bytes_written);
		       
static GIOError g_io_beos_write(GIOChannel *channel, 
				gchar     *buf, 
				guint      count,
				guint     *bytes_written);
static GIOError g_io_beos_seek (GIOChannel *channel,
				gint      offset, 
				GSeekType type);
static void g_io_beos_close    (GIOChannel *channel);
static void g_io_beos_free     (GIOChannel *channel);
static guint  g_io_beos_fd_add_watch (GIOChannel      *channel,
				   gint           priority,
				   GIOCondition   condition,
				   GIOFunc        func,
				   gpointer       user_data,
				   GDestroyNotify notify);

/* socket definitions!!! */

static GIOError g_io_beos_recv (GIOChannel *channel, 
		       gchar     *buf, 
		       guint      count,
		       guint     *bytes_written);
		       
static GIOError g_io_beos_send(GIOChannel *channel, 
				gchar     *buf, 
				guint      count,
				guint     *bytes_written);
static GIOError g_io_beos_no_seek (GIOChannel *channel,
				gint      offset, 
				GSeekType type);
static void g_io_beos_sock_close    (GIOChannel *channel);
static guint  g_io_beos_add_watch (GIOChannel      *channel,
				   gint           priority,
				   GIOCondition   condition,
				   GIOFunc        func,
				   gpointer       user_data,
				   GDestroyNotify notify);
static guint  g_io_beos_fd_add_watch (GIOChannel      *channel,
				   gint           priority,
				   GIOCondition   condition,
				   GIOFunc        func,
				   gpointer       user_data,
				   GDestroyNotify notify);

/* common definitions!!! */

static gboolean g_io_beos_prepare  (gpointer source_data, 
				    GTimeVal *current_time,
				    gint *timeout,
				    gpointer user_data);
static gboolean g_io_beos_check    (gpointer source_data,
				    GTimeVal *current_time,
				    gpointer user_data);
static gboolean g_io_beos_fd_dispatch (gpointer source_data,
				    GTimeVal *current_time,
				    gpointer user_data);
static gboolean g_io_beos_dispatch (gpointer source_data,
				    GTimeVal *current_time,
				    gpointer user_data);
static void g_io_beos_destroy  (gpointer source_data);
static void g_io_beos_fd_destroy  (gpointer source_data);

GSourceFuncs beos_fd_watch_funcs = {
  g_io_beos_prepare,
  g_io_beos_check,
  g_io_beos_fd_dispatch,
  g_io_beos_fd_destroy
};

GSourceFuncs beos_watch_funcs = {
  g_io_beos_prepare,
  g_io_beos_check,
  g_io_beos_dispatch,
  g_io_beos_destroy
};

GIOFuncs beos_channel_fd_funcs = {
  g_io_beos_read,
  g_io_beos_write,
  g_io_beos_seek,
  g_io_beos_close,
  g_io_beos_fd_add_watch,
  g_io_beos_free,
};

GIOFuncs beos_channel_sock_funcs = {
  g_io_beos_recv,
  g_io_beos_send,
  g_io_beos_no_seek,
  g_io_beos_sock_close,
  g_io_beos_add_watch,
  g_io_beos_free,
};

typedef struct {
	int 		fd;
	int 		signal;
	GIOBeosChannel *chan;
} fd_thread_data;

static port_id message_port = -1;

/* Uncomment this to allow for more debugging messages
 * for pipes/fd's
 */
// #define DEBUG_FDS
	
/* This thread routine is started to watch over an fd.
 * Basically we sit in a blocking read call and then send a message to
 * a port that the system uses for watching these threads.
 * At present sending an 'X' means error and G_IO_ERR will be set,
 * 'R' means there's something to read so G_IO_IN will be set.
 *
 * The gio-test program didin't terminate correctly when EOF was
 * sent via returning 0 bytes, so we currently treat this as an
 * error condition.  Don't know is this is correct or is totally bogus
 * and only time will truly tell!
 */
static int32
fd_watch_thread(void *data)
{
	fd_thread_data *fdd = (fd_thread_data*)data;
	GIOBeosChannel *bc = fdd->chan;
	int signal_sock = fdd->signal;
	int myfd = fdd->fd;
	char *byte = (char*)malloc(sizeof(char*));
	int len = 1, rv;
    bool run = true;
    
	/* this will block! */
    while (run){
      #ifdef DEBUG_FDS
        g_print("presently waiting = %d\n", bc->waiting);
      #endif
      while (bc->waiting == 1){snooze(1000);}
      acquire_sem(bc->mutex);
      #ifdef DEBUG_FDS
        g_print("trying to read a byte from fd %d = ", myfd);
      #endif
      rv = read(myfd, byte, len);
      #ifdef DEBUG_FDS
        g_print("%d\n", rv);
      #endif
      /* Hmm, not sure if this isn't totally bogus,
       * but to signal the end we treat EOF as an error and return
       * as such. 
       */
      if (rv <= 0){
        write_port(message_port, signal_sock, "X" , 1);        
        run = false;
        release_sem(bc->mutex);
      } else {
        bc->byte = byte;
        bc->waiting = 1;
        #ifdef DEBUG_FDS
          g_print("got a byte on fd %d (%#X) - signalling...\n", myfd, *byte);
	    #endif
        release_sem(bc->mutex);
      	write_port(message_port, signal_sock, "R", 1);
      }
	}
	free(byte);
	/* our work is done...so exit! */
}

static gboolean 
g_io_beos_prepare  (gpointer source_data, 
		    GTimeVal *current_time,
		    gint     *timeout,
		    gpointer user_data)
{
  *timeout = -1;
  return FALSE;
}

static gboolean 
g_io_beos_check    (gpointer source_data,
		    GTimeVal *current_time,
		    gpointer user_data)
{
  GIOBeosWatch *data = source_data;

  return (data->pollfd.revents & data->condition);
}

static gboolean
g_io_beos_dispatch (gpointer source_data, 
		    GTimeVal *current_time,
		    gpointer user_data)

{
  GIOBeosWatch *data = source_data;

  return (*data->callback)(data->channel,
			   data->pollfd.revents & data->condition,
			   user_data);
}

static gboolean
g_io_beos_fd_dispatch (gpointer source_data, 
		    GTimeVal *current_time,
		    gpointer user_data)

{
  GIOBeosWatch *data = source_data;
  
  return (*data->callback)(data->channel,
			   G_IO_IN,
			   user_data);
}

static void 
g_io_beos_destroy (gpointer source_data)
{
  GIOBeosWatch *data = source_data;

  g_main_remove_poll (&data->pollfd);
  g_io_channel_unref (data->channel);
  g_free (data);
}

static void 
g_io_beos_fd_destroy (gpointer source_data)
{
  GIOBeosWatch *data = source_data;
g_print("g_io_beos_fd_destroy\n");
  closesocket(data->sock);
  kill_thread(data->tid);
  g_main_remove_poll (&data->pollfd);  
  g_io_channel_unref (data->channel);
  g_free (data);
}

static GIOError 
g_io_beos_read (GIOChannel *channel, 
		gchar     *buf, 
		guint      count,
		guint     *bytes_read)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;
  gint result = 0;
  gint bytes = 0;
  
  acquire_sem(beos_channel->mutex);
  if (beos_channel->waiting == 1)
    {
    *buf = *beos_channel->byte;
    bytes = 1;
    beos_channel->waiting = 0;
    }
    
  result = read (beos_channel->fd, buf+bytes, count-bytes);
  release_sem(beos_channel->mutex);
  
  if (result < 0 && bytes == 1)
    result = 1;
  else if (bytes == 1)
    result++;

#ifdef DEBUG_FDS
  g_print("g_io_beos_read : fd %d : %d bytes\n", beos_channel->fd, result);
#endif

  if (result < 0)
    {
      *bytes_read = 0;
      switch (errno)
	{
	case EINVAL:
	  return G_IO_ERROR_INVAL;
	case EAGAIN:
	case EINTR:
	  return G_IO_ERROR_AGAIN;
	default:
	  return G_IO_ERROR_UNKNOWN;
	}
    }
  else
    {
      *bytes_read = result;
      return G_IO_ERROR_NONE;
    }
}
		       
static GIOError 
g_io_beos_write(GIOChannel *channel, 
		gchar     *buf, 
		guint      count,
		guint     *bytes_written)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;
  gint result;
  g_print("g_io_beos_write\n");
  result = write (beos_channel->fd, buf, count);

  if (result < 0)
    {
      *bytes_written = 0;
      switch (errno)
	{
	case EINVAL:
	  return G_IO_ERROR_INVAL;
	case EAGAIN:
	case EINTR:
	  return G_IO_ERROR_AGAIN;
	default:
	  return G_IO_ERROR_UNKNOWN;
	}
    }
  else
    {
      *bytes_written = result;
      return G_IO_ERROR_NONE;
    }
}

static GIOError 
g_io_beos_seek (GIOChannel *channel,
		gint      offset, 
		GSeekType type)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;
  int whence;
  off_t result;

  switch (type)
    {
    case G_SEEK_SET:
      whence = SEEK_SET;
      break;
    case G_SEEK_CUR:
      whence = SEEK_CUR;
      break;
    case G_SEEK_END:
      whence = SEEK_END;
      break;
    default:
      g_warning ("g_io_beos_seek: unknown seek type");
      return G_IO_ERROR_UNKNOWN;
    }

  result = lseek (beos_channel->fd, offset, whence);

  if (result < 0)
    {
      switch (errno)
	{
	case EINVAL:
	  return G_IO_ERROR_INVAL;
	default:
	  return G_IO_ERROR_UNKNOWN;
	}
    }
  else
    return G_IO_ERROR_NONE;
}


static void 
g_io_beos_close (GIOChannel *channel)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;

  close (beos_channel->fd);
}

static void 
g_io_beos_free (GIOChannel *channel)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;

  g_free (beos_channel);
}

static guint 
g_io_beos_add_watch (GIOChannel    *channel,
		     gint           priority,
		     GIOCondition   condition,
		     GIOFunc        func,
		     gpointer       user_data,
		     GDestroyNotify notify)
{
  GIOBeosWatch *watch = g_new (GIOBeosWatch, 1);
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;
  
  watch->channel = channel;
  g_io_channel_ref (channel);

  watch->callback = func;
  watch->condition = condition;

  watch->pollfd.fd = beos_channel->fd;
  watch->pollfd.events = condition;

  g_main_add_poll (&watch->pollfd, priority);

  return g_source_add (priority, TRUE, &beos_watch_funcs, watch, user_data, notify);
}

static guint 
g_io_beos_fd_add_watch (GIOChannel    *channel,
		     gint           priority,
		     GIOCondition   condition,
		     GIOFunc        func,
		     gpointer       user_data,
		     GDestroyNotify notify)
{
  GIOBeosWatch *watch = g_new (GIOBeosWatch, 1);
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;
  /* malloc it as it'll be freed later */
  fd_thread_data *fdd = (fd_thread_data*)malloc(sizeof(fd_thread_data*));
  /* this is the sock that we'll use to give us a unique number
   * for the thread.  when we close it on destroying it'll be re-allocated,
   * and this might cause us problems but hopefully the watch that uses the
   * unique id will have died before that!
   */
  int sock = socket(AF_INET, SOCK_DGRAM, 0);

  /* the global watch port for fd threads */
  if(message_port == -1){
    message_port = create_port(32, "glib_message_port");
  }
  
  watch->channel = channel;
  g_io_channel_ref (channel);

  watch->callback = func;
  watch->condition = condition;
  /* keep a record of the unique id.  should really rename this! */
  watch->sock = sock + 1000;

  /* the watch pollfd will be given the unique id as it's fd */
  watch->pollfd.fd = sock + 1000;
  watch->pollfd.events = condition;
  g_main_add_poll (&watch->pollfd, priority);

  /* fill in the data structure for our watch thread */ 
  fdd->fd = beos_channel->fd; /* the fd we're going to read from!!! */
  fdd->chan = beos_channel;
  fdd->signal = sock + 1000;
  
  watch->tid = spawn_thread(fd_watch_thread, "fd_watch_thread", B_NORMAL_PRIORITY,
    (void*)fdd);  
  if (watch->tid > 0)
    resume_thread(watch->tid);
  
  return g_source_add (priority, TRUE, &beos_watch_funcs, watch, user_data, notify);
}

static GIOError 
g_io_beos_recv (GIOChannel *channel, 
		      gchar      *buf, 
		      guint       count,
		      guint      *bytes_read)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *) channel;
  gint result;

  result = recv (beos_channel->fd, buf, count, 0);
  if (result < 0)
    {
      *bytes_read = 0;
      switch (errno)
	    {
	      case EINVAL:
	        return G_IO_ERROR_INVAL;
	      case EWOULDBLOCK:
	      case EINTR:
	        return G_IO_ERROR_AGAIN;
	      default:
	        return G_IO_ERROR_UNKNOWN;
	    }
      }
    else
      {
        *bytes_read = result;
        return G_IO_ERROR_NONE;
      }
}
		       
static GIOError 
g_io_beos_send(GIOChannel *channel, 
		      gchar      *buf, 
		      guint       count,
		      guint      *bytes_written)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *) channel;
  gint result;

  result = send (beos_channel->fd, buf, count, 0);
      
  if (result < 0)
    {
      *bytes_written = 0;
      switch (errno)
	{
	case EINVAL:
	  return G_IO_ERROR_INVAL;
	case EWOULDBLOCK:
	case EINTR:
	  return G_IO_ERROR_AGAIN;
	default:
	  return G_IO_ERROR_UNKNOWN;
	}
    }
  else
    {
      *bytes_written = result;
      return G_IO_ERROR_NONE;
    }
}

static void 
g_io_beos_sock_close (GIOChannel *channel)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *) channel;

  closesocket (beos_channel->fd - 1000);
  return;
}

/*
static guint 
g_io_beos_sock_add_watch (GIOChannel    *channel,
			   gint           priority,
			   GIOCondition   condition,
			   GIOFunc        func,
			   gpointer       user_data,
			   GDestroyNotify notify)
{
  GIOBeosWatch *watch = g_new (GIOBeosWatch, 1);
  GIOBeosChannel *beos_channel = (GIOBeosChannel *) channel;
  
  watch->channel = channel;
  g_io_channel_ref (channel);

  watch->callback = func;
  watch->condition = condition;

  watch->pollfd.fd = beos_channel->fd;
  watch->pollfd.events = condition;

  g_main_add_poll (&watch->pollfd, priority);

  return g_source_add (priority, TRUE, &beos_watch_funcs, watch, user_data, notify);
}
*/

GIOChannel *
g_io_channel_unix_new (gint fd)
{
  GIOBeosChannel *beos_channel = g_new (GIOBeosChannel, 1);
  GIOChannel *channel = (GIOChannel *)beos_channel;

  g_io_channel_init (channel);
  channel->funcs = &beos_channel_fd_funcs;
  
  beos_channel->mutex = create_sem(1,"GIOChannel mutex");
  beos_channel->fd = fd;
  beos_channel->type = G_IO_FILE_DESC;
  return channel;
}

GIOChannel *
g_io_channel_beos_new_socket (int socket)
{
  GIOBeosChannel *beos_channel = g_new (GIOBeosChannel, 1);
  GIOChannel *channel = (GIOChannel *) beos_channel;

  g_io_channel_init (channel);
  channel->funcs = &beos_channel_sock_funcs;
  beos_channel->fd = socket;
  beos_channel->type = G_IO_SOCKET;

  return channel;
}


static GIOError 
g_io_beos_no_seek (GIOChannel *channel,
		    gint      offset, 
		    GSeekType type)
{
  g_warning ("g_io_beos_no_seek: unseekable IO channel type");
  return G_IO_ERROR_UNKNOWN;
}


gint
g_io_channel_unix_get_fd (GIOChannel *channel)
{
  GIOBeosChannel *beos_channel = (GIOBeosChannel *)channel;
  return beos_channel->fd;
}
